#pragma once

#include <stdint.h>
#include <avr/pgmspace.h>
#include "i2c/utils.hpp"


namespace Devices::Sh1106
{
    constexpr uint8_t const COLUMNS_N = (128);
    constexpr uint8_t const COLUMN_OFFSET = (2);
    constexpr uint8_t const PAGES_N = (8);
    constexpr uint8_t const PAGE_HEIGHT = (8);

    constexpr uint8_t const WIDTH = (COLUMNS_N);
    constexpr uint8_t const HEIGHT = (PAGES_N * PAGE_HEIGHT);

    constexpr uint8_t const CTRL_COMMAND = (0x00);
    constexpr uint8_t const CTRL_DATA = (0x40);
    constexpr uint8_t const CTRL_N_BYTES = (0x00);
    constexpr uint8_t const CTRL_SINGLE_BYTE = (0x80);

    constexpr uint8_t const CMD_SET_LOW_COLUMN(uint8_t const low) { return (0x00 | ((low) & 0x0f)); }
    constexpr uint8_t const CMD_SET_HIGH_COLUMN(uint8_t const hi) { return (0x10 | ((hi) & 0x0f)); }
    constexpr uint8_t const CMD_SET_PUMP_VOLTAGE_VALUE(uint8_t const value) { return (0x30 | ((value) & 0x03)); }
    constexpr uint8_t const CMD_SET_DISPLAY_START_LINE(uint8_t const line) { return (0x40 | ((line) & 0x3f)); }
    constexpr uint8_t const CMD_SET_CONTRAST = (0x81);
    constexpr uint8_t const CMD_SET_SEGMENT_REMAP(bool const is_enabled) { return (0xA0 | ((is_enabled) & 0x01)); }
    constexpr uint8_t const CMD_SET_ENTIRE_DISPLAY_ON(bool const is_on) { return (0xA4 | ((is_on) & 0x01)); }
    constexpr uint8_t const CMD_SET_INVERSED(bool const is_inversed) { return (0xA6 | ((is_inversed) & 0x01)); }
    constexpr uint8_t const CMD_SET_MULTIPLEX_RATIO = (0xA8);
    constexpr uint8_t const CMD_SET_DCDC_CONTROL_MODE = (0xAD);
    constexpr uint8_t const CMD_SET_DCDC_ON(bool const is_on) { return (0x8A | ((is_on) & 0x01)); }
    constexpr uint8_t const CMD_SET_DISPLAY_ON(bool const is_on) { return (0xAE | ((is_on) & 0x01)); }
    constexpr uint8_t const CMD_SET_PAGE_ADDRESS(uint8_t const page) { return (0xB0 | ((page) & 0x07)); }
    constexpr uint8_t const CMD_SET_COM_SCAN_DIRECTION_INC = (0xC0);
    constexpr uint8_t const CMD_SET_COM_SCAN_DIRECTION_DEC = (0xC8);
    constexpr uint8_t const CMD_SET_DISPLAY_OFFSET = (0xD3);
    constexpr uint8_t const CMD_SET_CLOCK_DIVIDE_FREQUENCY = (0xD5);
    constexpr uint8_t const CMD_SET_PRECHARGE_PERIOD = (0xD9);
    constexpr uint8_t const CMD_SET_COM_PINS_HW_CONF = (0xDA);
    constexpr uint8_t const CMD_SET_VCOMH_DESELECT_LEVEL = (0xDB);
    constexpr uint8_t const CMD_READ_MODIFY_WRITE_START = (0xE0);
    constexpr uint8_t const CMD_READ_MODIFY_WRITE_END = (0xEE);
    constexpr uint8_t const CMD_NOP = (0xE3);

    extern uint8_t const INIT_SEQUENCE[24] PROGMEM;

    class DrawCommand 
    {
        private:

            uint8_t width;
            uint8_t address;
            uint8_t ctrl1;
            uint8_t cmd_high_column;
            uint8_t ctrl2;
            uint8_t cmd_low_column;
            uint8_t ctrl3;
            uint8_t cmd_page;
            uint8_t ctrl4;

        public:

            uint8_t gddram_data[16];

            DrawCommand(uint8_t const column, uint8_t const page, uint8_t const width)
                : width(width)
                , ctrl1(CTRL_COMMAND | CTRL_SINGLE_BYTE)
                , cmd_high_column(CMD_SET_HIGH_COLUMN((column + COLUMN_OFFSET) >> 4))
                , ctrl2(CTRL_COMMAND | CTRL_SINGLE_BYTE)
                , cmd_low_column(CMD_SET_LOW_COLUMN((column + COLUMN_OFFSET) & 0x0f))
                , ctrl3(CTRL_COMMAND | CTRL_SINGLE_BYTE)
                , cmd_page(CMD_SET_PAGE_ADDRESS(page))
                , ctrl4(CTRL_DATA | CTRL_N_BYTES)
            {
            }

            void SetAddress(uint8_t const address)
            {
                this->address = address;
            }

            uint8_t GetSize() const
            {
                return this->width + 7;
            }

            uint8_t const * GetBuffer() const
            {
                return &ctrl1;
            }
    };


    template <class I2cImplT, uint8_t SA0>
    class Driver
    {
        public:

            using DrawCommand = Sh1106::DrawCommand;

            constexpr static uint8_t const ADDRESS = 0x78 | (SA0 << 1);

            static bool Init()
            {
                I2cImplT::Init();
                return Configure();
            }

            static bool Draw(DrawCommand & command)
            {
                command.SetAddress(ADDRESS);

                bool result = true;
                result = result && I2cImplT::Start(ADDRESS, I2c::Direction::Transmit);
                result = result && I2cImplT::Transmit(command.GetSize(), command.GetBuffer());
                result = result && I2cImplT::Stop();
                return result;
            }

            static bool MoveTo(uint8_t const column, uint8_t const page)
            {
                uint8_t positioning[] =
                {
                    CTRL_COMMAND | CTRL_SINGLE_BYTE,
                    CMD_SET_HIGH_COLUMN((column + COLUMN_OFFSET) >> 4),
                    CTRL_COMMAND | CTRL_SINGLE_BYTE,
                    CMD_SET_LOW_COLUMN((column + COLUMN_OFFSET) & 0x0f),
                    CTRL_COMMAND | CTRL_SINGLE_BYTE,
                    CMD_SET_PAGE_ADDRESS(page)
                };

                bool result = true;
                result = result && I2cImplT::Start(ADDRESS, I2c::Direction::Transmit);
                result = result && I2cImplT::Transmit(sizeof(positioning), positioning);
                result = result && I2cImplT::Stop();
                return result;
            }

            static bool SendPixelData(uint8_t const width, uint8_t const * data)
            {
                bool result = true;
                result = result && I2cImplT::Start(ADDRESS, I2c::Direction::Transmit);
                result = result && I2cImplT::Transmit(CTRL_DATA | CTRL_N_BYTES);
                result = result && I2cImplT::Transmit(width, data);
                result = result && I2cImplT::Stop();
                return result;
            }

            static bool Draw(uint8_t const column, uint8_t const page, uint8_t const width, uint8_t const * data)
            {
                bool result = true;
                result = result && MoveTo(column, page);
                result = result && SendPixelData(width, data);
                return result;
            }

            static bool Clear()
            {
                bool result = true;

                for (uint8_t i = 0; i < PAGES_N; ++i) 
                {
                    result = result && MoveTo(0, i);
                    result = result && I2cImplT::Start(ADDRESS, I2c::Direction::Transmit);
                    result = result && I2cImplT::Transmit(CTRL_DATA | CTRL_N_BYTES);

                    for (uint8_t i = 0; i < COLUMNS_N; ++i)
                    {
                        result = result && I2cImplT::Transmit(0);
                    }
                    
                    result = result && I2cImplT::Stop();
                }

                return result;
            }

        private:

            static bool Configure()
            {
                bool result = true;
                result = result && I2cImplT::Start(ADDRESS, I2c::Direction::Transmit);
                result = result && I2cImplT::TransmitProgmem(sizeof(INIT_SEQUENCE), INIT_SEQUENCE);
                result = result && I2cImplT::Stop();
                return result;
            }
    };
}